import importlib

import numpy as np
import pytest

import quimb.tensor as qtn

requires_autograd = pytest.mark.skipif(
    importlib.util.find_spec("autograd") is None,
    reason="autograd not installed",
)


@pytest.mark.parametrize("method", ("auto", "dense", "overlap"))
@pytest.mark.parametrize(
    "normalized",
    (
        True,
        False,
        "squared",
        "infidelity",
        "infidelity_sqrt",
    ),
)
def test_tensor_network_distance(method, normalized):
    n = 6
    A = qtn.TN_rand_reg(n=n, reg=3, D=2, phys_dim=2, dtype=complex)
    Ad = A.to_dense([f"k{i}" for i in range(n)])
    B = qtn.TN_rand_reg(n=6, reg=3, D=2, phys_dim=2, dtype=complex)
    Bd = B.to_dense([f"k{i}" for i in range(n)])
    d1 = np.linalg.norm(Ad - Bd)
    d2 = A.distance(B, method=method, normalized=normalized)
    if normalized:
        assert 0 <= d2 <= 2
    else:
        assert d1 == pytest.approx(d2)


@pytest.mark.parametrize(
    "opts",
    [
        dict(method="als", dense_solve=False),
        dict(method="als", dense_solve=False, solver="lgmres"),
        dict(
            method="als",
            dense_solve=True,
            enforce_pos=False,
            solver_dense="lstsq",
        ),
        dict(method="als", dense_solve=True, enforce_pos=True),
        dict(method="tree"),
        pytest.param(
            dict(method="autodiff", distance_method="dense"),
            marks=requires_autograd,
        ),
        pytest.param(
            dict(method="autodiff", distance_method="overlap"),
            marks=requires_autograd,
        ),
    ],
)
@pytest.mark.parametrize("dtype", ("float64", "complex128"))
def test_fit_mps(opts, dtype):
    k1 = qtn.MPS_rand_state(5, 3, seed=666, dtype=dtype)
    k2 = qtn.MPS_rand_state(5, 3, seed=667, dtype=dtype)
    assert k1.distance_normalized(k2) > 1e-3
    k1.fit_(k2, progbar=True, **dict(opts))
    assert k1.distance_normalized(k2) < 1e-3


@pytest.mark.parametrize(
    "opts",
    [
        dict(method="als", dense_solve=False),
        dict(method="als", dense_solve=False, solver="lgmres"),
        dict(
            method="als",
            dense_solve=True,
            enforce_pos=False,
            solver_dense="lstsq",
        ),
        dict(method="als", dense_solve=True, enforce_pos=True),
        pytest.param(
            dict(method="autodiff", distance_method="dense"),
            marks=requires_autograd,
        ),
        pytest.param(
            dict(method="autodiff", distance_method="overlap"),
            marks=requires_autograd,
        ),
    ],
)
@pytest.mark.parametrize("dtype", ("float64", "complex128"))
def test_fit_rand_reg(opts, dtype):
    r1 = qtn.TN_rand_reg(5, 4, D=2, seed=666, phys_dim=2, dtype=dtype)
    k2 = qtn.MPS_rand_state(5, 3, seed=667, dtype=dtype)
    assert r1.distance(k2) > 1e-3
    r1.fit_(k2, progbar=True, **dict(opts))
    assert r1.distance(k2) < 1e-3


@pytest.mark.parametrize(
    "opts",
    [
        dict(method="als", dense_solve=False),
        dict(method="als", dense_solve=False, solver="lgmres"),
        dict(
            method="als",
            dense_solve=True,
            enforce_pos=False,
            solver_dense="lstsq",
        ),
        dict(method="als", dense_solve=True, enforce_pos=True),
        dict(method="tree"),
        pytest.param(
            dict(method="autodiff", distance_method="dense"),
            marks=requires_autograd,
        ),
        pytest.param(
            dict(method="autodiff", distance_method="overlap"),
            marks=requires_autograd,
        ),
    ],
)
@pytest.mark.parametrize("dtype", ("float64", "complex128"))
def test_fit_partial_tags(opts, dtype):
    k1 = qtn.MPS_rand_state(5, 3, seed=666, dtype=dtype)
    k2 = qtn.MPS_rand_state(5, 3, seed=667, dtype=dtype)
    d0 = k1.distance(k2)
    tags = ["I0", "I2", "I4"]
    k1f = k1.fit(k2, tol=1e-3, tags=tags, progbar=True, **dict(opts))
    assert k1f.distance(k2) < d0
    if opts["method"] != "tree":
        assert (k1f[0] - k1[0]).norm() > 1e-12
        assert (k1f[1] - k1[1]).norm() < 1e-12
        assert (k1f[2] - k1[2]).norm() > 1e-12
        assert (k1f[3] - k1[3]).norm() < 1e-12
        assert (k1f[4] - k1[4]).norm() > 1e-12


def test_hyper_distance():
    tna = qtn.HTN_rand(10, 3, 2, 2, 2, 2, 2, seed=0, dtype="complex128")
    tnb = qtn.HTN_rand(10, 3, 2, 2, 2, 2, 2, seed=1, dtype="complex128")
    oix = [f"k{i}" for i in range(4)]
    xa = tna.to_dense(oix)
    xb = tnb.to_dense(oix)
    assert tna.norm(output_inds=oix) == pytest.approx(np.linalg.norm(xa))
    assert tnb.norm(output_inds=oix) == pytest.approx(np.linalg.norm(xb))
    assert tnb.overlap(tna, output_inds=oix) == pytest.approx(np.vdot(xa, xb))
    assert tna.distance(tnb, output_inds=oix) == pytest.approx(
        np.linalg.norm(xa - xb)
    )



import pytest
from numpy.testing import assert_allclose

import quimb.operator as qop


def assert_all_matrices_match(
    sob: qop.SparseOperatorBuilder,
    extras=(),
):
    Ad = sob.build_dense()
    As = sob.build_sparse_matrix()
    assert_allclose(Ad, As.toarray(), atol=1e-10)
    Am = sob.build_mpo().to_dense()
    assert_allclose(Ad, Am, atol=1e-10)
    Ae = sob.build_matrix_ikron()
    assert_allclose(Ad, Ae, atol=1e-10)
    for extra in extras:
        assert_allclose(Ae, extra, atol=1e-10)
    return Ae


@pytest.mark.parametrize("n", [1, 2, 3, 5])
@pytest.mark.parametrize("m", [1, 3, 10])
@pytest.mark.parametrize("k", [1, 2, 3])
@pytest.mark.parametrize("kmin", [0, None])
@pytest.mark.parametrize("seed", [0, 1, 2])
def test_rand_operator_matrices(n, m, k, kmin, seed):
    if k > n:
        pytest.skip("k > n, skipping test")
    sob = qop.rand_operator(n, m, k, kmin=kmin, seed=seed, ops="xyz+-n")
    assert sob.nsites == n
    assert sob.nterms <= m  # can be less due to repeated terms
    A0 = assert_all_matrices_match(sob)
    sob.pauli_decompose()
    assert_all_matrices_match(sob, extras=[A0])
    sob.pauli_decompose()
    assert_all_matrices_match(sob, extras=[A0])
    sob.pauli_decompose(use_zx=True)
    assert_all_matrices_match(sob, extras=[A0])



import quimb.tensor as qtn
from quimb.utils import oset

from .bp_common import (
    BeliefPropagationCommon,
    combine_local_contractions,
    create_lazy_community_edge_map,
)


class L1BP(BeliefPropagationCommon):
    """Lazy 1-norm belief propagation. BP is run between groups of tensors
    defined by ``site_tags``. The message updates are lazy contractions.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on.
    site_tags : sequence of str, optional
        The tags identifying the sites in ``tn``, each tag forms a region,
        which should not overlap. If the tensor network is structured, then
        these are inferred automatically.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    contract_every : int, optional
        If not None, 'contract' (via BP) the tensor network every
        ``contract_every`` iterations. The resulting values are stored in
        ``zvals`` at corresponding points ``zval_its``.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.
    """

    def __init__(
        self,
        tn,
        site_tags=None,
        *,
        damping=0.0,
        update="sequential",
        normalize=None,
        distance=None,
        local_convergence=True,
        optimize="auto-hq",
        message_init_function=None,
        contract_every=None,
        inplace=False,
        **contract_opts,
    ):
        super().__init__(
            tn,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            contract_every=contract_every,
            inplace=inplace,
        )

        self.local_convergence = local_convergence
        self.optimize = optimize
        self.contract_opts = contract_opts

        if site_tags is None:
            self.site_tags = tuple(tn.site_tags)
        else:
            self.site_tags = tuple(site_tags)

        (
            self.edges,
            self.neighbors,
            self.local_tns,
            self.touch_map,
        ) = create_lazy_community_edge_map(tn, site_tags)
        self.touched = oset()

        # for each meta bond create initial messages
        self.messages = {}
        for pair, bix in self.edges.items():
            # compute leftwards and rightwards messages
            for i, j in (sorted(pair), sorted(pair, reverse=True)):
                tn_i = self.local_tns[i]
                # initial message just sums over dangling bonds

                if message_init_function is None:
                    tm = tn_i.contract(
                        all,
                        output_inds=bix,
                        optimize=self.optimize,
                        drop_tags=True,
                        **self.contract_opts,
                    )
                    # normalize
                    tm.modify(apply=self._normalize_fn)
                else:
                    shape = tuple(tn_i.ind_size(ix) for ix in bix)
                    tm = qtn.Tensor(
                        data=message_init_function(shape),
                        inds=bix,
                    )

                self.messages[i, j] = tm

        # compute the contractions
        self.contraction_tns = {}
        for pair, bix in self.edges.items():
            # for each meta bond compute left and right contractions
            for i, j in (sorted(pair), sorted(pair, reverse=True)):
                tn_i = self.local_tns[i].copy()
                # attach incoming messages to dangling bonds
                tks = [
                    self.messages[k, i] for k in self.neighbors[i] if k != j
                ]
                # virtual so we can modify messages tensors inplace
                tn_i_to_j = qtn.TensorNetwork((tn_i, *tks), virtual=True)
                self.contraction_tns[i, j] = tn_i_to_j

    def iterate(self, tol=5e-6):
        if (not self.local_convergence) or (not self.touched):
            # assume if asked to iterate that we want to check all messages
            self.touched.update(
                pair for edge in self.edges for pair in (edge, edge[::-1])
            )

        ncheck = len(self.touched)
        nconv = 0
        max_mdiff = -1.0
        new_touched = oset()

        def _compute_m(key):
            i, j = key
            bix = self.edges[(i, j) if i < j else (j, i)]
            tn_i_to_j = self.contraction_tns[i, j]
            tm_new = tn_i_to_j.contract(
                all,
                output_inds=bix,
                optimize=self.optimize,
                **self.contract_opts,
            )
            return self._normalize_fn(tm_new.data)

        def _update_m(key, data):
            nonlocal nconv, max_mdiff

            tm = self.messages[key]

            # pre-damp distance
            mdiff = self._distance_fn(data, tm.data)

            if self.damping:
                data = self._damping_fn(data, tm.data)

            # # post-damp distance
            # mdiff = self._distance_fn(data, tm.data)

            if mdiff > tol:
                # mark touching messages for update
                new_touched.update(self.touch_map[key])
            else:
                nconv += 1

            max_mdiff = max(max_mdiff, mdiff)
            tm.modify(data=data)

        if self.update == "parallel":
            new_data = {}
            # compute all new messages
            while self.touched:
                key = self.touched.pop()
                new_data[key] = _compute_m(key)
            # insert all new messages
            for key, data in new_data.items():
                _update_m(key, data)

        elif self.update == "sequential":
            # compute each new message and immediately re-insert it
            while self.touched:
                key = self.touched.pop()
                data = _compute_m(key)
                _update_m(key, data)

        self.touched = new_touched
        return {
            "nconv": nconv,
            "ncheck": ncheck,
            "max_mdiff": max_mdiff,
        }

    def contract(self, strip_exponent=False, check_zero=True):
        zvals = []
        for site, tn_ic in self.local_tns.items():
            if site in self.neighbors:
                tval = qtn.tensor_contract(
                    *tn_ic,
                    *(self.messages[k, site] for k in self.neighbors[site]),
                    optimize=self.optimize,
                    **self.contract_opts,
                )
            else:
                # site exists but has no neighbors
                tval = tn_ic.contract(
                    all,
                    output_inds=(),
                    optimize=self.optimize,
                    **self.contract_opts,
                )
            zvals.append((tval, 1))

        for i, j in self.edges:
            mval = qtn.tensor_contract(
                self.messages[i, j],
                self.messages[j, i],
                optimize=self.optimize,
                **self.contract_opts,
            )
            # power / counting factor is -1 for messages
            zvals.append((mval, -1))

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
        )

    def normalize_message_pairs(self):
        """Normalize all messages such that for each bond `<m_i|m_j> = 1` and
        `<m_i|m_i> = <m_j|m_j>` (but in general != 1).
        """
        for i, j in self.edges:
            tmi = self.messages[i, j]
            tmj = self.messages[j, i]
            nij = abs(tmi @ tmj) ** 0.5
            nii = (tmi @ tmi) ** 0.25
            njj = (tmj @ tmj) ** 0.25
            tmi /= nij * nii / njj
            tmj /= nij * njj / nii


def contract_l1bp(
    tn,
    max_iterations=1000,
    tol=5e-6,
    site_tags=None,
    damping=0.0,
    update="sequential",
    diis=False,
    local_convergence=True,
    optimize="auto-hq",
    strip_exponent=False,
    info=None,
    progbar=False,
    **contract_opts,
):
    """Estimate the contraction of ``tn`` using lazy 1-norm belief propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to contract.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    site_tags : sequence of str, optional
        The tags identifying the sites in ``tn``, each tag forms a region. If
        the tensor network is structured, then these are inferred
        automatically.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    update : {'parallel', 'sequential'}, optional
        Whether to update all messages in parallel or sequentially.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    progbar : bool, optional
        Whether to show a progress bar.
    strip_exponent : bool, optional
        Whether to strip the exponent from the final result. If ``True``
        then the returned result is ``(mantissa, exponent)``.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.
    """
    bp = L1BP(
        tn,
        site_tags=site_tags,
        damping=damping,
        local_convergence=local_convergence,
        update=update,
        optimize=optimize,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        diis=diis,
        info=info,
        progbar=progbar,
    )
    return bp.contract(
        strip_exponent=strip_exponent,
    )



import pytest

import quimb as qu
import quimb.tensor as qtn
import quimb.tensor.belief_propagation as qbp


@pytest.mark.parametrize("local_convergence", [False, True])
def test_contract_tree_exact(local_convergence):
    tn = qtn.TN_rand_tree(20, 3)
    Z = tn.contract()
    info = {}
    Z_bp = qbp.contract_d1bp(
        tn, info=info, local_convergence=local_convergence, progbar=True
    )
    assert info["converged"]
    assert Z == pytest.approx(Z_bp, rel=1e-12)


@pytest.mark.parametrize("damping", [0.0, 0.1])
@pytest.mark.parametrize("diis", [False, True])
def test_contract_normal(damping, diis):
    tn = qtn.TN2D_from_fill_fn(lambda s: qu.randn(s, dist="uniform"), 6, 6, 2)
    Z = tn.contract()
    info = {}
    Z_bp = qbp.contract_d1bp(
        tn, damping=damping, diis=diis, info=info, progbar=True
    )
    assert info["converged"]
    assert Z == pytest.approx(Z_bp, rel=1e-1)


def test_get_gauged_tn():
    tn = qtn.TN2D_from_fill_fn(lambda s: qu.randn(s, dist="uniform"), 6, 6, 2)
    Z = tn.contract()
    bp = qbp.D1BP(tn)
    bp.run()
    Zbp = bp.contract()
    assert Z == pytest.approx(Zbp, rel=1e-1)
    tn_gauged = bp.get_gauged_tn()
    Zg = qu.prod(array.item(0) for array in tn_gauged.arrays)
    assert Z == pytest.approx(Zg, rel=1e-1)



import contextlib
import functools
import itertools
import operator

import autoray as ar

import quimb.tensor as qtn
from quimb.utils import oset

from .bp_common import (
    BeliefPropagationCommon,
    combine_local_contractions,
    normalize_message_pair,
    process_loop_series_expansion_weights,
)
from .regions import gen_region_counts


def _parse_global_gloops(tn, gloops=None):
    if isinstance(gloops, int):
        max_size = gloops
        gloops = None
    else:
        max_size = None

    if gloops is None:
        gloops = tuple(tn.gen_gloops(max_size=max_size))
    else:
        gloops = tuple(gloops)

    return gloops


class D2BP(BeliefPropagationCommon):
    """Dense (as in one tensor per site) 2-norm (as in for wavefunctions and
    operators) belief propagation. Allows messages reuse. This version assumes
    no hyper indices (i.e. a standard PEPS like tensor network).

    Potential use cases for D2BP and a PEPS like tensor network are:

        - globally compressing it from bond dimension ``D`` to ``D'``
        - eagerly applying gates and locally compressing back to ``D``
        - sampling configurations
        - estimating the norm of the tensor network


    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to form the 2-norm of and run BP on.
    messages : dict[(str, int), array_like], optional
        The initial messages to use, effectively defaults to all ones if not
        specified.
    output_inds : set[str], optional
        The indices to consider as output (dangling) indices of the tn.
        Computed automatically if not specified.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    contract_every : int, optional
        If not None, 'contract' (via BP) the tensor network every
        ``contract_every`` iterations. The resulting values are stored in
        ``zvals`` at corresponding points ``zval_its``.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.
    """

    def __init__(
        self,
        tn,
        *,
        messages=None,
        output_inds=None,
        optimize="auto-hq",
        damping=0.0,
        update="sequential",
        normalize=None,
        distance=None,
        local_convergence=True,
        contract_every=None,
        inplace=False,
        **contract_opts,
    ):
        super().__init__(
            tn=tn,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            contract_every=contract_every,
            inplace=inplace,
        )

        self.contract_opts = contract_opts
        self.contract_opts.setdefault("optimize", optimize)
        self.local_convergence = local_convergence

        if output_inds is None:
            self.output_inds = set(self.tn.outer_inds())
        else:
            self.output_inds = set(output_inds)

        if messages is None:
            self.messages = {}
        else:
            self.messages = messages

        # record which messages touch each others, for efficient updates
        self.touch_map = {}
        self.touched = oset()
        self.exprs = {}

        # populate any messages
        for ix, tids in self.tn.ind_map.items():
            if ix in self.output_inds:
                continue

            tida, tidb = tids
            jx = ix + "*"
            ta, tb = self.tn._tids_get(tida, tidb)

            for tid, t, t_in in ((tida, ta, tb), (tidb, tb, ta)):
                this_touchmap = []
                for nx in t.inds:
                    if nx in self.output_inds or nx == ix:
                        continue
                    # where this message will be sent on to
                    (tidn,) = (n for n in self.tn.ind_map[nx] if n != tid)
                    this_touchmap.append((nx, tidn))
                self.touch_map[ix, tid] = this_touchmap

                if (ix, tid) not in self.messages:
                    m = (t_in.reindex({ix: jx}).conj_() @ t_in).data
                    self.messages[ix, tid] = self._normalize_fn(m)

        # for efficiency setup all the contraction expressions ahead of time
        for ix, tids in self.tn.ind_map.items():
            if ix not in self.output_inds:
                self.build_expr(ix)

    def build_expr(self, ix):
        from quimb.tensor.contraction import array_contract_expression

        tids = self.tn.ind_map[ix]

        for tida, tidb in (sorted(tids), sorted(tids, reverse=True)):
            ta = self.tn.tensor_map[tida]
            kix = ta.inds
            bix = tuple(i if i in self.output_inds else i + "*" for i in kix)
            inputs = [kix, bix]
            data = [ta.data, ta.data.conj()]
            shapes = [ta.shape, ta.shape]
            for i in kix:
                if (i != ix) and i not in self.output_inds:
                    inputs.append((i + "*", i))
                    data.append((i, tida))
                    shapes.append(self.messages[i, tida].shape)

            expr = array_contract_expression(
                inputs=inputs,
                output=(ix + "*", ix),
                shapes=shapes,
                **self.contract_opts,
            )
            self.exprs[ix, tidb] = expr, data

    def update_touched_from_tids(self, *tids):
        """Specify that the messages for the given ``tids`` have changed."""
        for tid in tids:
            t = self.tn.tensor_map[tid]
            for ix in t.inds:
                if ix in self.output_inds:
                    continue
                (ntid,) = (n for n in self.tn.ind_map[ix] if n != tid)
                self.touched.add((ix, ntid))

    def update_touched_from_tags(self, tags, which="any"):
        """Specify that the messages for the messages touching ``tags`` have
        changed.
        """
        tids = self.tn._get_tids_from_tags(tags, which)
        self.update_touched_from_tids(*tids)

    def update_touched_from_inds(self, inds, which="any"):
        """Specify that the messages for the messages touching ``inds`` have
        changed.
        """
        tids = self.tn._get_tids_from_inds(inds, which)
        self.update_touched_from_tids(*tids)

    def iterate(self, tol=5e-6):
        """Perform a single iteration of dense 2-norm belief propagation."""

        if (not self.local_convergence) or (not self.touched):
            # assume if asked to iterate that we want to check all messages
            self.touched.update(self.exprs.keys())

        ncheck = len(self.touched)
        nconv = 0
        max_mdiff = -1.0
        new_touched = oset()

        def _compute_m(key):
            expr, data = self.exprs[key]
            m = expr(*data[:2], *(self.messages[mkey] for mkey in data[2:]))
            # enforce hermiticity and normalize
            return self._normalize_fn(m + ar.dag(m))

        def _update_m(key, new_m):
            nonlocal nconv, max_mdiff

            old_m = self.messages[key]

            # pre-damp distance
            mdiff = self._distance_fn(old_m, new_m)

            if self.damping:
                new_m = self._damping_fn(old_m, new_m)

            # # post-damp distance
            # mdiff = self._distance_fn(old_m, new_m)

            if mdiff > tol:
                # mark touching messages for update
                new_touched.update(self.touch_map[key])
            else:
                nconv += 1
            max_mdiff = max(max_mdiff, mdiff)
            self.messages[key] = new_m

        if self.update == "parallel":
            new_messages = {}
            # compute all new messages
            while self.touched:
                key = self.touched.pop()
                new_messages[key] = _compute_m(key)
            # insert all new messages
            for key, new_m in new_messages.items():
                _update_m(key, new_m)

        elif self.update == "sequential":
            # compute each new message and immediately re-insert it
            while self.touched:
                key = self.touched.pop()
                new_m = _compute_m(key)
                _update_m(key, new_m)

        self.touched = new_touched

        return {
            "nconv": nconv,
            "ncheck": ncheck,
            "max_mdiff": max_mdiff,
        }

    def compute_marginal(self, ind):
        """Compute the marginal for the index ``ind``."""
        (tid,) = self.tn.ind_map[ind]
        t = self.tn.tensor_map[tid]

        arrays = [t.data, ar.do("conj", t.data)]
        k_input = []
        b_input = []
        m_inputs = []
        for j, jx in enumerate(t.inds, 1):
            k_input.append(j)

            if jx == ind:
                # output index -> take diagonal
                output = (j,)
                b_input.append(j)
            else:
                try:
                    # partial trace with message
                    m = self.messages[jx, tid]
                    arrays.append(m)
                    b_input.append(-j)
                    m_inputs.append((-j, j))
                except KeyError:
                    # direct partial trace
                    b_input.append(j)

        p = qtn.array_contract(
            arrays,
            inputs=(tuple(k_input), tuple(b_input), *m_inputs),
            output=output,
            **self.contract_opts,
        )
        p = ar.do("real", p)
        return p / ar.do("sum", p)

    def normalize_message_pairs(self):
        """Normalize a pair of messages such that `<mi|mj> = 1` and
        `<mi|mi> = <mj|mj>` (but in general != 1).
        """
        _reshape = ar.get_lib_fn(self.backend, "reshape")

        for ix, tids in self.tn.ind_map.items():
            if len(tids) != 2:
                continue
            tida, tidb = tids
            ml = self.messages[ix, tida]
            mr = self.messages[ix, tidb]

            nml, nmr = normalize_message_pair(
                _reshape(ml, (-1,)),
                _reshape(mr, (-1,)),
            )

            self.messages[ix, tida] = _reshape(nml, ml.shape)
            self.messages[ix, tidb] = _reshape(nmr, mr.shape)

    def local_tensor_contract(self, tid):
        """Contract the local region of the tensor at ``tid``."""
        t = self.tn.tensor_map[tid]
        arrays = [t.data, ar.do("conj", t.data)]
        k_input = []
        b_input = []
        m_inputs = []
        for i, ix in enumerate(t.inds, 1):
            k_input.append(i)
            if ix in self.output_inds:
                b_input.append(i)
            else:
                b_input.append(-i)
                m_inputs.append((-i, i))
                arrays.append(self.messages[ix, tid])

        inputs = (tuple(k_input), tuple(b_input), *m_inputs)
        output = ()
        return qtn.array_contract(arrays, inputs, output, **self.contract_opts)

    def normalize_tensors(self, strip_exponent=True):
        """Normalize the tensors in the tensor network such that their 2-norm
        is 1. If ``strip_exponent`` is ``True`` then accrue the phase and
        exponent (log10) into the ``sign`` and ``exponent`` attributes of the
        D2BP object (the default), contract methods can then reinsert these
        factors when returning the final result.
        """
        for tid, t in self.tn.tensor_map.items():
            tval = self.local_tensor_contract(tid)
            tabs = ar.do("abs", tval)
            tsgn = tval / tabs
            tlog = ar.do("log10", tabs)
            t /= (tsgn * tabs) ** 0.5
            if strip_exponent:
                self.sign = tsgn * self.sign
                self.exponent = tlog + self.exponent

    def contract(
        self,
        strip_exponent=False,
        check_zero=True,
    ):
        """Estimate the total contraction, i.e. the 2-norm.

        Parameters
        ----------
        strip_exponent : bool, optional
            Whether to strip the exponent from the final result. If ``True``
            then the returned result is ``(mantissa, exponent)``.

        Returns
        -------
        scalar or (scalar, float)
        """
        zvals = []

        for tid in self.tn.tensor_map:
            tval = self.local_tensor_contract(tid)
            zvals.append((tval, 1))

        for ix, tids in self.tn.ind_map.items():
            if ix in self.output_inds:
                continue
            tida, tidb = tids
            ml = self.messages[ix, tidb]
            mr = self.messages[ix, tida]
            mval = qtn.array_contract(
                (ml, mr), ((1, 2), (1, 2)), (), **self.contract_opts
            )
            # counting factor is -1 i.e. divide by the message
            zvals.append((mval, -1))

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            mantissa=self.sign,
            exponent=self.exponent,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
        )

    def get_cluster_excited(
        self,
        tids=None,
        partial_trace_map=(),
        exclude=(),
    ):
        """Get the local norm tensor network for ``tids`` with BP messages
        inserted on the boundary and excitation projectors inserted on the
        inner bonds. See arxiv.org/abs/2409.03108 for more details.

        Parameters
        ----------
        tids : iterable of hashable
            The tensor ids to include in the cluster.
        partial_trace_map : dict[str, str], optional
            A remapping of ket indices to bra indices to perform an effective
            partial trace.
        exclude : iterable of str, optional
            A set of bond indices to exclude from inserting excitation
            projectors on, e.g. when forming a reduced density matrix.

        Returns
        -------
        TensorNetwork
        """
        stn = self.tn._select_tids(tids)

        kixmaps = {tid: {} for tid in stn.tensor_map}
        bixmaps = {tid: {} for tid in stn.tensor_map}
        exc_ixs = {}
        bms = []
        ems = []

        for ix, tids in stn.ind_map.items():
            if ix in self.output_inds:
                # physical traced index
                if ix in partial_trace_map:
                    (tid,) = tids
                    bixmaps[tid][ix] = partial_trace_map[ix]

            elif ix in exclude:
                # excluded bond -> simply rename bra
                bix = qtn.rand_uuid()
                for tid in tids:
                    bixmaps[tid][ix] = bix

            elif ix in stn._inner_inds:
                # internal index
                for tid in tids:
                    kix = qtn.rand_uuid()
                    bix = qtn.rand_uuid()
                    kixmaps[tid][ix] = kix
                    bixmaps[tid][ix] = bix
                    # store labels for excitation projector
                    exc_ixs.setdefault(ix, {})[tid] = (kix, bix)
                ems.append((ix, tids))

            else:
                # boundary index
                (tid,) = tids
                kix = qtn.rand_uuid()
                bix = qtn.rand_uuid()
                kixmaps[tid][ix] = kix
                bixmaps[tid][ix] = bix
                bms.append((ix, tid))

        tn = qtn.TensorNetwork()

        # add bra and ket tensors
        for tid in stn.tensor_map:
            t = stn.tensor_map[tid]
            tn |= t.reindex(kixmaps[tid])
            tn |= t.conj().reindex(bixmaps[tid])

        # add boundary message tensors
        for ix, tid in bms:
            data = self.messages[ix, tid]
            inds = (kixmaps[tid][ix], bixmaps[tid][ix])
            tn |= qtn.Tensor(data, inds)

        # add inner exitation projector message tensors
        with ar.backend_like(self.backend):
            for ix, tids in ems:
                tidl, tidr = tids
                ml = self.messages[ix, tidl]
                mr = self.messages[ix, tidr]

                # form outer product
                p0 = ar.do("einsum", "i,j->ij", ml.reshape(-1), mr.reshape(-1))
                # subtract from identity
                pe = ar.do("eye", ar.do("shape", p0)[0]) - p0
                # reshape back into 4-tensor
                pe = ar.do(
                    "reshape",
                    pe,
                    ar.do("shape", ml) + ar.do("shape", mr),
                )
                inds = (*exc_ixs[ix][tidl], *exc_ixs[ix][tidr])
                tn |= qtn.Tensor(pe, inds)

        return tn

    def contract_loop_series_expansion(
        self,
        gloops=None,
        multi_excitation_correct=True,
        tol_correction=1e-12,
        maxiter_correction=100,
        strip_exponent=False,
        optimize="auto-hq",
        **contract_opts,
    ):
        """Contract the norm of the tensor network using the same procedure as
        in https://arxiv.org/abs/2409.03108 - "Loop Series Expansions for
        Tensor Networks".

        Parameters
        ----------
        gloops : int or iterable of tuples, optional
            The gloop sizes to use. If an integer, then generate all gloop
            sizes up to this size. If a tuple, then use these gloops.
        multi_excitation_correct : bool, optional
            Whether to use the multi-excitation correction. If ``True``, then
            the free energy is refined iteratively until self consistent.
        tol_correction : float, optional
            The tolerance for the multi-excitation correction.
        maxiter_correction : int, optional
            The maximum number of iterations for the multi-excitation
            correction.
        strip_exponent : bool, optional
            Whether to strip the exponent from the final result. If ``True``
            then the returned result is ``(mantissa, exponent)``.
        optimize : str or PathOptimizer, optional
            The path optimizer to use when contracting the messages.
        contract_opts
            Other options supplied to ``TensorNetwork.contract``.
        """
        self.normalize_message_pairs()
        # accrues BP estimate into self.sign and self.exponent
        self.normalize_tensors()

        gloops = _parse_global_gloops(self.tn, gloops)

        weights = {}
        for gloop in gloops:
            # get local tensor network with boundary
            # messages and excitation projectors inserted
            etn = self.get_cluster_excited(gloop)
            # contract it to get local weight!
            weights[tuple(gloop)] = etn.contract(
                optimize=optimize, **contract_opts
            )

        return process_loop_series_expansion_weights(
            weights,
            mantissa=self.sign,
            exponent=self.exponent,
            multi_excitation_correct=multi_excitation_correct,
            tol_correction=tol_correction,
            maxiter_correction=maxiter_correction,
            strip_exponent=strip_exponent,
        )

    def partial_trace_loop_series_expansion(
        self,
        where,
        gloops=None,
        normalized=True,
        grow_from="alldangle",
        strict_size=False,
        multi_excitation_correct=True,
        optimize="auto-hq",
        **contract_opts,
    ):
        """Compute the reduced density matrix for the sites specified by
        ``where`` using the loop series expansion method from
        https://arxiv.org/abs/2409.03108 - "Loop Series Expansions for Tensor
        Networks".

        Parameters
        ----------
        where : sequence[hashable]
            The sites to from the reduced density matrix of.
        gloops : int or iterable of tuples, optional
            The generalized loops to use, or an integer to automatically
            generate all up to a certain size. If none use the smallest non-
            trivial size.
        normalized : bool, optional
            Whether to normalize the final density matrix.
        grow_from : {'alldangle', 'all', 'any'}, optional
            How to grow the generalized loops from the specified ``where``:

            - 'alldangle': clusters up to max size, where target sites are
              allowed to dangle.
            - 'all': clusters where loop, up to max size, has to include *all*
              target sites.
            - 'any': clusters where loop, up to max size, can include *any* of
              the target sites. Remaining target sites are added as extras.

            By default 'alldangle'.
        strict_size : bool, optional
            Whether to enforce the maximum size of the generalized loops, only
            relevant for `grow_from="any"`.
        multi_excitation_correct : bool, optional
            Whether to use the multi-excitation correction. If ``True``, then
            the free energy is refined iteratively until self consistent.
        optimize : str or PathOptimizer, optional
            The path optimizer to use when contracting the messages.
        contract_opts
            Other options supplied to ``TensorNetwork.contract``.
        """
        self.normalize_message_pairs()
        self.normalize_tensors()

        tags = [self.tn.site_tag(coo) for coo in where]
        tids = self.tn._get_tids_from_tags(tags, "any")

        # get a mapping of ket indices to bra indices on target sites
        kix = [self.tn.site_ind(coo) for coo in where]
        bix = [qtn.rand_uuid() for _ in where]
        partial_trace_map = dict(zip(kix, bix))
        output_inds = (*kix, *bix)

        # generate the generalized loops relevant to the target sites
        gloops = self.tn.get_local_gloops(
            tids=tids,
            gloops=gloops,
            grow_from=grow_from,
            strict_size=strict_size,
        )
        # the base (BP) region, including target sites only
        r0 = frozenset(tids)

        # get internal indices of the base region to exclude
        # from inserting excited space projectors on
        inner_bonds = self.tn._select_tids(tids).inner_inds()

        # get loop excited reduced density matrices
        rho_es = {}
        for gloop in gloops:
            etn = self.get_cluster_excited(
                gloop, exclude=inner_bonds, partial_trace_map=partial_trace_map
            )
            rho_e = etn.contract(
                output_inds=output_inds, optimize=optimize, **contract_opts
            )
            rho_e = rho_e.to_dense(kix, bix)

            if (normalized == "local") and gloop != r0:
                rho_e /= 1 + ar.do("trace", rho_e)

            rho_es[gloop] = rho_e

        if multi_excitation_correct:
            # trace of each density matrix is its corresponding
            # loops contribution to the norm free energy
            weights = {
                gloop: ar.do("trace", rho_e) for gloop, rho_e in rho_es.items()
            }
            # remove the BP contribution (= minimal region)
            weights.pop(r0)
            # compute exponential suppresion factors
            corrections = process_loop_series_expansion_weights(
                weights, return_all=True
            )
            # add back in the BP contribution
            corrections[r0] = 1.0

            # weighted sum
            rho = functools.reduce(
                operator.add,
                (
                    rho_e * corrections[gloop]
                    for gloop, rho_e in rho_es.items()
                ),
            )
        else:
            rho = functools.reduce(operator.add, rho_es.values())

        if normalized:
            rho /= ar.do("trace", rho)
        elif (self.sign, self.exponent) != (1.0, 0.0):
            # have been accrued into by normalize_tensors most likely
            rho *= self.sign * 10**self.exponent

        return rho

    def contract_gloop_expand(
        self,
        gloops=None,
        autocomplete=True,
        optimize="auto-hq",
        strip_exponent=False,
        check_zero=True,
        info=None,
        progbar=False,
        **contract_opts,
    ):
        self.normalize_message_pairs()

        gloops = _parse_global_gloops(self.tn, gloops)

        if info is None:
            info = {}
        info.setdefault("contractions", {})
        contractions = info["contractions"]

        region_counts = gen_region_counts(
            itertools.chain(gloops, ((tid,) for tid in self.tn.tensor_map)),
            autocomplete=autocomplete,
        )

        if progbar:
            import tqdm

            region_counts = tqdm.tqdm(region_counts)

        zvals = []
        for region, counting_factor in region_counts:
            try:
                zr = contractions[region]
            except KeyError:
                tnr = self.get_cluster_norm(region)
                zr = tnr.contract(optimize=optimize, **contract_opts)
                contractions[region] = zr
            zvals.append((zr, counting_factor))

        return combine_local_contractions(
            zvals,
            mantissa=self.sign,
            exponent=self.exponent,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
        )

    def compress(
        self,
        max_bond,
        cutoff=0.0,
        cutoff_mode=4,
        renorm=0,
        inplace=False,
    ):
        """Compress the initial tensor network using the current messages."""
        tn = self.tn if inplace else self.tn.copy()

        for ix, tids in tn.ind_map.items():
            if len(tids) != 2:
                continue
            tida, tidb = tids

            # messages are left and right factors squared already
            ta = tn.tensor_map[tida]
            dm = ta.ind_size(ix)
            dl = ta.size // dm
            ml = self.messages[ix, tidb]
            Rl = qtn.decomp.squared_op_to_reduced_factor(
                ml, dl, dm, right=True
            )

            tb = tn.tensor_map[tidb]
            dr = tb.size // dm
            mr = self.messages[ix, tida].T
            Rr = qtn.decomp.squared_op_to_reduced_factor(
                mr, dm, dr, right=False
            )

            # compute the compressors
            Pl, Pr = qtn.decomp.compute_oblique_projectors(
                Rl,
                Rr,
                max_bond=max_bond,
                cutoff=cutoff,
                cutoff_mode=cutoff_mode,
                renorm=renorm,
            )

            # contract the compressors into the tensors
            tn.tensor_map[tida].gate_(Pl.T, ix)
            tn.tensor_map[tidb].gate_(Pr, ix)

            # update messages with projections
            if inplace:
                new_Ra = Rl @ Pl
                new_Rb = Pr @ Rr
                self.messages[ix, tidb] = ar.dag(new_Ra) @ new_Ra
                self.messages[ix, tida] = new_Rb @ ar.dag(new_Rb)

        return tn

    def gauge_insert(self, tn, smudge=1e-12):
        """Insert the sqrt of messages on the boundary of a part of the main BP
        TN.

        Parameters
        ----------
        tn : TensorNetwork
            The tensor network to insert the messages into.
        smudge : float, optional
            Smudge factor to avoid numerical issues, the eigenvalues of the
            messages are clipped to be at least the largest eigenvalue times
            this factor.

        Returns
        -------
        list[tuple[Tensor, str, array_like]]
            The sequence of tensors, indices and inverse gauges to apply to
            reverse the gauges applied.
        """
        outer = []

        _eigh = ar.get_lib_fn(self.backend, "linalg.eigh")
        _clip = ar.get_lib_fn(self.backend, "clip")
        _sqrt = ar.get_lib_fn(self.backend, "sqrt")

        for ix in tn.outer_inds():
            # get the tensor and dangling index
            (tid,) = tn.ind_map[ix]
            try:
                m = self.messages[ix, tid]
            except KeyError:
                # could be phsyical index or not generated yet
                continue
            t = tn.tensor_map[tid]

            # compute the 'square root' of the message
            s2, W = _eigh(m)
            s2 = _clip(s2, s2[-1] * smudge, None)
            s = _sqrt(s2)
            msqrt = qtn.decomp.ldmul(s, ar.dag(W))
            msqrt_inv = qtn.decomp.rddiv(W, s)
            t.gate_(msqrt, ix)
            outer.append((t, ix, msqrt_inv))

        return outer

    @contextlib.contextmanager
    def gauge_temp(self, tn, ungauge_outer=True):
        """Context manager to temporarily gauge a tensor network, presumably a
        subnetwork of the main BP network, using the current messages, and then
        un-gauge it afterwards.

        Parameters
        ----------
        tn : TensorNetwork
            The tensor network to gauge.
        ungauge_outer : bool, optional
            Whether to un-gauge the outer indices of the tensor network.
        """
        outer = self.gauge_insert(tn)
        try:
            yield outer
        finally:
            if ungauge_outer:
                for t, ix, msqrt_inv in outer:
                    t.gate_(msqrt_inv, ix)

    def gate_(
        self,
        G,
        where,
        max_bond=None,
        cutoff=0.0,
        cutoff_mode="rsum2",
        renorm=0,
        tn=None,
        **gate_opts,
    ):
        """Apply a gate to the tensor network at the specified sites, using
        the current messages to gauge the tensors.
        """
        if len(where) == 1:
            # single site gate
            self.tn.gate_(G, where, contract=True)
            return

        gate_opts.setdefault("contract", "reduce-split")

        if tn is None:
            tn = self.tn
        site_tags = tuple(map(tn.site_tag, where))
        tn_where = tn.select_any(site_tags)

        with self.gauge_temp(tn_where):
            # contract and split the gate
            tn_where.gate_(
                G,
                where,
                max_bond=max_bond,
                cutoff=cutoff,
                cutoff_mode=cutoff_mode,
                renorm=renorm,
                **gate_opts,
            )

            # update the messages for this bond
            taga, tagb = site_tags
            (tida,) = tn._get_tids_from_tags(taga)
            (tidb,) = tn._get_tids_from_tags(tagb)
            ta = tn.tensor_map[tida]
            tb = tn.tensor_map[tidb]
            lix, (ix,), rix = qtn.group_inds(ta, tb)

            # make use of the fact that we already have gauged tensors
            A = ta.to_dense(lix, (ix,))
            B = tb.to_dense((ix,), rix)
            ma = ar.dag(A) @ A
            mb = B @ ar.dag(B)

            shape_changed = self.messages[ix, tidb].shape != ma.shape

            self.messages[ix, tidb] = ma
            self.messages[ix, tida] = mb

            # mark the sites as touched
            self.update_touched_from_tids(tida, tidb)
            if shape_changed:
                # rebuild the contraction expressions if shapes changed
                for cix in (*lix, ix, *rix):
                    if cix not in self.output_inds:
                        self.build_expr(cix)

    def get_cluster_norm(
        self,
        tids,
        partial_trace_map=(),
    ):
        """Get the local norm tensor network for ``tids`` with BP messages
        inserted on the boundary. Optionally open some physical indices up to
        perform an effective partial trace.

        Parameters
        ----------
        tids : iterable of hashable
            The tensor ids to include in the cluster.
        partial_trace_map : dict[str, str], optional
            A remapping of ket indices to bra indices to perform an effective
            partial trace.

        Returns
        -------
        TensorNetwork
        """
        k = self.tn._select_tids(tids, virtual=False)
        b = k.conj()

        if partial_trace_map:
            # open up the bra indices
            b.reindex_(partial_trace_map)

        for ix in k.outer_inds():
            if (ix not in partial_trace_map) and (ix not in self.output_inds):
                # dangling index -> absorb message into ket
                (tid,) = k.ind_map[ix]
                t = k.tensor_map[tid]
                t.gate_(self.messages[ix, tid], ix)

        return b | k

    def partial_trace(
        self,
        where,
        normalized=True,
        tids_region=None,
        get="matrix",
        bra_ind_id=None,
        optimize="auto-hq",
        **contract_opts,
    ):
        """Get the reduced density matrix for the sites specified by ``where``,
        with the remaining network approximated by messages on the boundary.

        Parameters
        ----------
        where : sequence[hashable]
            The sites to from the reduced density matrix of.
        get : {'tn', 'tensor', 'array', 'matrix'}, optional
            The type of object to return. If 'tn', return the uncontracted
            tensor network object. If 'tensor', return the labelled density
            operator as a `Tensor`. If 'array', return the unfused raw array
            with 2 * len(where) dimensions. If 'matrix', fuse the ket and bra
            indices and return this 2D matrix.
        bra_ind_id : str, optional
            If ``get="tn"``, how to label the bra indices. If None, use the
            default based on the current site_ind_id.
        optimize : str or PathOptimizer, optional
            The path optimizer to use when contracting the tensor network.
        contract_opts
            Other options supplied to ``TensorNetwork.contract``.

        Returns
        -------
        TensorNetwork or Tensor or array
        """
        # get a mapping of ket indices to bra indices on target sites
        if bra_ind_id is None:
            bra_ind_id = "b" + self.tn.site_ind_id[1:]
        bra_ind_starmap = bra_ind_id.count("{}") > 1
        kix = [self.tn.site_ind(coo) for coo in where]
        if bra_ind_starmap:
            bix = [bra_ind_id.format(*coo) for coo in where]
        else:
            bix = [bra_ind_id.format(coo) for coo in where]
        output_inds = (*kix, *bix)
        partial_trace_map = dict(zip(kix, bix))

        # get target region
        tags = [self.tn.site_tag(coo) for coo in where]

        if tids_region is None:
            tids_region = self.tn._get_tids_from_tags(tags, "any")
        tn = self.get_cluster_norm(
            tids_region, partial_trace_map=partial_trace_map
        )

        if get == "tn":
            return tn

        t = tn.contract(
            output_inds=output_inds, optimize=optimize, **contract_opts
        )

        if normalized:
            t /= t.trace(kix, bix)

        if get == "tensor":
            return t
        elif get == "array":
            return t.data
        elif get == "matrix":
            return t.to_dense(kix, bix)
        else:
            raise ValueError(f"Unknown get option: {get}")

    def partial_trace_gloop_expand(
        self,
        where,
        gloops=None,
        combine="sum",
        normalized=True,
        grow_from="alldangle",
        strict_size=False,
        optimize="auto-hq",
        **contract_opts,
    ):
        """Compute a reduced density matrix for the sites specified by
        ``where`` using the generalized loop cluster expansion.

        Parameters
        ----------
        where : sequence[hashable]
            The sites to from the reduced density matrix of.
        gloops : int or iterable of tuples, optional
            The generalized loops to use, or an integer to automatically
            generate all up to a certain size. If none use the smallest non-
            trivial size.
        combine : {'sum', 'prod'}, optional
            How to combine the contributions from each generalized loop. If
            'sum', use coefficient weighted addition. If 'prod', use power
            weighted multiplication.
        normalized : bool or {"local", "separate"}, optional
            Whether to normalize the density matrix. If True or "local",
            normalize each cluster density matrix by its trace. If "separate",
            normalize the final density matrix by its trace (usually less
            accurate). If False, do not normalize.
        grow_from : {'alldangle', 'all', 'any'}, optional
            How to grow the generalized loops from the specified ``where``:

            - 'alldangle': clusters up to max size, where target sites are
              allowed to dangle.
            - 'all': clusters where loop, up to max size, has to include *all*
              target sites.
            - 'any': clusters where loop, up to max size, can include *any* of
              the target sites. Remaining target sites are added as extras.

            By default 'alldangle'.
        strict_size : bool, optional
            Whether to enforce the maximum size of the generalized loops, only
            relevant for `grow_from="any"`.
        optimize : str or PathOptimizer, optional
            The path optimizer to use when contracting the tensor network.
        contract_opts
            Other options supplied to ``TensorNetwork.contract``.
        """
        tags = [self.tn.site_tag(coo) for coo in where]
        tids = self.tn._get_tids_from_tags(tags, "any")

        if normalized is True:
            normalized = "local"

        gloops = self.tn.get_local_gloops(
            tids=tids,
            gloops=gloops,
            grow_from=grow_from,
            strict_size=strict_size,
        )

        rhos = []
        for region, cr in gen_region_counts(gloops):
            rho_r = self.partial_trace(
                where,
                tids_region=region,
                normalized=False,
                get="matrix",
                optimize=optimize,
                **contract_opts,
            )

            if normalized == "local":
                rho_r /= ar.do("trace", rho_r)

            rhos.append((rho_r, cr))

        if combine == "sum":
            rho = functools.reduce(
                operator.add, (cr * rho_r for rho_r, cr in rhos)
            )
        elif combine == "prod":
            rho = functools.reduce(
                operator.mul, (rho_r**cr for rho_r, cr in rhos)
            )
        else:
            raise ValueError(f"Unknown combine option: {combine}")

        if (normalized == "separate") or (normalized and combine == "prod"):
            rho /= ar.do("trace", rho)

        if (not normalized) and ((self.sign, self.exponent) != (1.0, 0.0)):
            # have been accrued into by normalize_tensors most likely
            rho *= self.sign * 10**self.exponent

        return rho


def contract_d2bp(
    tn,
    *,
    messages=None,
    output_inds=None,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    diis=False,
    update="sequential",
    normalize=None,
    distance=None,
    tol_abs=None,
    tol_rolling_diff=None,
    local_convergence=True,
    optimize="auto-hq",
    strip_exponent=False,
    check_zero=True,
    info=None,
    progbar=False,
    **contract_opts,
):
    """Estimate the norm squared of ``tn`` using dense 2-norm belief
    propagation (no hyper indices).

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to form the 2-norm of and run BP on.
    messages : dict[(str, int), array_like], optional
        The initial messages to use, effectively defaults to all ones if not
        specified.
    output_inds : set[str], optional
        The indices to consider as output (dangling) indices of the tn.
        Computed automatically if not specified.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    strip_exponent : bool, optional
        Whether to return the mantissa and exponent separately.
    check_zero : bool, optional
        Whether to check for zero values and return zero early.
    info : dict, optional
        If supplied, the following information will be added to it:
        ``converged`` (bool), ``iterations`` (int), ``max_mdiff`` (float),
        ``rolling_abs_mean_diff`` (float).
    progbar : bool, optional
        Whether to show a progress bar.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.

    Returns
    -------
    scalar or (scalar, float)
    """
    bp = D2BP(
        tn,
        messages=messages,
        output_inds=output_inds,
        optimize=optimize,
        local_convergence=local_convergence,
        damping=damping,
        update=update,
        normalize=normalize,
        distance=distance,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        diis=diis,
        tol=tol,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
        info=info,
        progbar=progbar,
    )
    return bp.contract(
        strip_exponent=strip_exponent,
        check_zero=check_zero,
    )


def compress_d2bp(
    tn,
    max_bond,
    cutoff=0.0,
    cutoff_mode="rsum2",
    renorm=0,
    messages=None,
    output_inds=None,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    diis=False,
    update="sequential",
    normalize=None,
    distance=None,
    tol_abs=None,
    tol_rolling_diff=None,
    local_convergence=True,
    optimize="auto-hq",
    inplace=False,
    info=None,
    progbar=False,
    **contract_opts,
):
    """Compress the tensor network ``tn`` using dense 2-norm belief
    propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to form the 2-norm of, run BP on and then compress.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        The cutoff to use when compressing.
    cutoff_mode : int, optional
        The cutoff mode to use when compressing.
    renorm : float, optional
        Whether to renormalize the singular values when compressing.
    messages : dict[(str, int), array_like], optional
        The initial messages to use, effectively defaults to all ones if not
        specified.
    output_inds : set[str], optional
        The indices to consider as output (dangling) indices of the tn.
        Computed automatically if not specified.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    inplace : bool, optional
        Whether to perform the compression inplace.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.

    Returns
    -------
    TensorNetwork
    """
    bp = D2BP(
        tn,
        messages=messages,
        output_inds=output_inds,
        optimize=optimize,
        damping=damping,
        update=update,
        normalize=normalize,
        distance=distance,
        local_convergence=local_convergence,
        inplace=inplace,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        diis=diis,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
        info=info,
        progbar=progbar,
    )
    return bp.compress(
        max_bond=max_bond,
        cutoff=cutoff,
        cutoff_mode=cutoff_mode,
        renorm=renorm,
        inplace=inplace,
    )


def sample_d2bp(
    tn,
    output_inds=None,
    messages=None,
    max_iterations=100,
    tol=1e-2,
    bias=None,
    seed=None,
    optimize="auto-hq",
    damping=0.0,
    diis=False,
    update="sequential",
    normalize=None,
    distance=None,
    tol_abs=None,
    tol_rolling_diff=None,
    local_convergence=True,
    progbar=False,
    **contract_opts,
):
    """Sample a configuration from ``tn`` using dense 2-norm belief
    propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to sample from.
    output_inds : set[str], optional
        Which indices to sample.
    messages : dict[(str, int), array_like], optional
        The initial messages to use, effectively defaults to all ones if not
        specified.
    max_iterations : int, optional
        The maximum number of iterations to perform, per marginal.
    tol : float, optional
        The convergence tolerance for messages.
    bias : float, optional
        Bias the sampling towards more locally likely bit-strings. This is
        done by raising the probability of each bit-string to this power.
    seed : int, optional
        A random seed for reproducibility.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    progbar : bool, optional
        Whether to show a progress bar.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.

    Returns
    -------
    config : dict[str, int]
        The sampled configuration, a mapping of output indices to values.
    tn_config : TensorNetwork
        The tensor network with the sampled configuration applied.
    omega : float
        The BP probability of the sampled configuration.
    """
    import numpy as np

    if output_inds is None:
        output_inds = tn.outer_inds()

    rng = np.random.default_rng(seed)
    config = {}
    omega = 1.0

    tn = tn.copy()
    bp = D2BP(
        tn,
        messages=messages,
        optimize=optimize,
        damping=damping,
        update=update,
        normalize=normalize,
        distance=distance,
        local_convergence=local_convergence,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        diis=diis,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
    )

    marginals = dict.fromkeys(output_inds)

    if progbar:
        import tqdm

        pbar = tqdm.tqdm(total=len(marginals))
    else:
        pbar = None

    while marginals:
        for ix in marginals:
            marginals[ix] = bp.compute_marginal(ix)

        ix, p = max(marginals.items(), key=lambda x: max(x[1]))
        p = ar.to_numpy(p)

        if bias is not None:
            # bias distribution towards more locally likely bit-strings
            p = p**bias
            p /= np.sum(p)

        v = rng.choice([0, 1], p=p)
        config[ix] = v
        del marginals[ix]

        tids = tuple(tn.ind_map[ix])
        tn.isel_({ix: v})

        omega *= p[v]
        if progbar:
            pbar.update(1)
            pbar.set_description(f"{ix}->{v}", refresh=False)

        bp = D2BP(
            tn,
            messages=messages,
            optimize=optimize,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            local_convergence=local_convergence,
            **contract_opts,
        )
        bp.update_touched_from_tids(*tids)
        bp.run(
            max_iterations=max_iterations,
            tol=tol,
            diis=diis,
            tol_abs=tol_abs,
            tol_rolling_diff=tol_rolling_diff,
        )

    if progbar:
        pbar.close()

    return config, tn, omega



def _compute_oblique_projectors_between_full_env_exact_tids(
    self,
    tida,
    tidb,
    max_bond=None,
    cutoff=1e-10,
    fit_steps=0,
    **fit_opts,
):
    from autoray import do
    from quimb.tensor import rand_uuid, Tensor

    k = self.copy()

    ta = k.tensor_map[tida]
    tb = k.tensor_map[tidb]

    # names for bonds
    (bix,) = ta.bonds(tb)
    ix_kl = rand_uuid()
    ix_kr = rand_uuid()
    ix_bl = rand_uuid()
    ix_br = rand_uuid()
    ebix = rand_uuid()

    # make norm network, but with ket and bra bond cut open
    k._cut_between_tids(tida, tidb, ix_kl, ix_kr)
    b = k.conj().reindex_({ix_kl: ix_bl, ix_kr: ix_br})
    ftn = k | b

    # contract full environment
    ft = ftn.contract(
        all,
        output_inds=[ix_kl, ix_kr, ix_bl, ix_br],
        optimize="auto-hq",
    )

    # factor positive environment
    ekt, _ = ft.split(
        left_inds=[ix_kl, ix_kr],
        right_inds=[ix_bl, ix_br],
        get="tensors",
        bond_ind=ebix,
        method="eigh",
    )

    ek = ft.trace([ix_bl], [ix_br]).to_dense([ix_kl], [ix_kr])
    U, _, VH = do("linalg.svd", ek)
    Pl = do("conj", U[:, :max_bond])
    Pr = do("conj", VH[:max_bond, :])

    # # compute left reduced factor
    # Rl = ekt.compute_reduced_factor("right", [ebix, ix_kr], [ix_kl])

    # # compute right reduced factor
    # Rr = ekt.compute_reduced_factor("left", [ix_kr], [ix_kl, ebix])

    # # compute compressed projectors
    # Pl, Pr = compute_oblique_projectors(
    #     Rl,
    #     Rr,
    #     max_bond=max_bond,
    #     cutoff=cutoff,
    # )

    if fit_steps:
        tPl = Tensor(Pl, inds=[ix_kl, bix], tags="Pl")
        tPr = Tensor(Pr, inds=[bix, ix_kr], tags="Pr")
        tn_fit = ekt | tPl | tPr
        tn_target = ekt.trace(ix_kl, ix_kr)
        tn_fit.fit_(
            tn_target,
            tags=["Pl", "Pr"],
            steps=fit_steps,
            **fit_opts,
        )
        Pl = tPl.data
        Pr = tPr.data

    return Pl, Pr


def _compress_between_full_env_exact_tids(
    self,
    tida,
    tidb,
    max_bond=None,
    cutoff=1e-10,
    fit_steps=0,
    inplace=False,
    **fit_opts,
):
    from quimb.tensor.tensor_core import tensor_make_single_bond

    tn = self if inplace else self.copy()

    ta = tn.tensor_map[tida]
    tb = tn.tensor_map[tidb]

    _, bix, _ = tensor_make_single_bond(ta, tb)

    Pl, Pr = tn._compute_oblique_projectors_between_full_env_exact_tids(
        tida,
        tidb,
        max_bond=max_bond,
        cutoff=cutoff,
        fit_steps=fit_steps,
        **fit_opts,
    )

    ta.gate_(Pl, bix, transposed=True)
    tb.gate_(Pr, bix)

    return tn


def _compress_between_cluster_env_tids(
    self,
    tida,
    tidb,
    max_bond=None,
    cutoff=1e-10,
    fit_steps=10,
    max_distance=1,
    fillin=False,
    gauges=None,
    power=1,
    smudge=1e-12,
    inplace=False,
    **fit_opts,
):
    tn = self if inplace else self.copy()
    k = tn._select_local_tids(
        [tida, tidb],
        max_distance=max_distance,
        fillin=fillin,
    )

    if gauges is not None:
        outer, inner = k.gauge_simple_insert(
            gauges,
            power=power,
            smudge=smudge,
        )
    else:
        outer = inner = None

    k._compress_between_full_env_exact_tids(
        tida,
        tidb,
        max_bond=max_bond,
        cutoff=cutoff,
        fit_steps=fit_steps,
        inplace=True,
        **fit_opts,
    )

    if gauges is not None:
        # undo outer gauge
        k.gauge_simple_remove(outer)
        # inner gauges are simply out of date and removed
        for _, ix, _ in inner:
            gauges.pop(ix)

        # regauge
        tn._gauge_local_tids(
            (tida, tidb),
            max_distance=max_distance,
            method="simple",
            gauges=gauges,
            power=power,
            smudge=smudge,
        )
        # tn.gauge_all_simple_(
        #     100,
        #     1e-9,
        #     gauges=gauges,
        #     touched_tids=(tida, tidb),
        # )

    return tn


def gate_cluster_environment(
    self,
    G,
    where,
    max_bond=None,
    cutoff=1e-10,
    fit_steps=10,
    max_distance=1,
    power=1,
    smudge=1e-12,
    gauges=None,
    inplace=False,
    **fit_opts,
):
    tn = self if inplace else self.copy()

    # apply gate without truncation
    if gauges is not None:
        tn.gate_simple_(
            G,
            where,
            gauges=gauges,
            max_bond=None,
            power=power,
            smudge=smudge,
        )
    else:
        tn.gate_(
            G,
            where,
            contract="reduce-split",
            max_bond=None,
        )

    tags = tuple(map(tn.site_tag, where))
    tida, tidb = tn._get_tids_from_tags(tags, "any")

    # compress corresponding bond using cluster environment
    return tn._compress_between_cluster_env_tids(
        tida,
        tidb,
        max_bond=max_bond,
        cutoff=cutoff,
        fit_steps=fit_steps,
        max_distance=max_distance,
        gauges=gauges,
        power=power,
        smudge=smudge,
        inplace=True,
        **fit_opts,
    )



"""Miscellaneous experimental functions."""

import functools

from quimb.tensor.tensor_core import (
    Tensor,
    _parse_split_opts,
    bonds,
    check_opt,
    dag,
    decomp,
    do,
    ensure_dict,
    group_inds,
    oset,
    rand_uuid,
    reshape,
    tensor_canonize_bond,
    tensor_compress_bond,
    tensor_contract,
    tensor_fuse_squeeze,
    tensor_make_single_bond,
    tensor_split,
)


def rand_reduce(self, ind, rand_fn=None, inplace=False):
    """Contract a random vector with ``ind``, removing it from this tensor.

    Parameters
    ----------
    ind : str
        The index to contract with.
    inplace : bool, optional
        Whether to perform the reduction inplace.

    Returns
    -------
    Tensor
    """
    t = self if inplace else self.copy()

    d = t.ind_size(ind)

    if rand_fn is None:
        r = do("random.normal", size=d, like=t.data)
        r /= do("linalg.norm", r)
    else:
        r = rand_fn(d)

    axis = t.inds.index(ind)
    new_inds = t.inds[:axis] + t.inds[axis + 1 :]
    t.modify(
        apply=lambda x: do("tensordot", x, r, axes=((axis,), (0,))),
        inds=new_inds,
    )
    return t


rand_reduce_ = functools.partialmethod(rand_reduce, inplace=True)


def _compress_between_local_reflect(
    self,
    tid1,
    tid2,
    method="svd",
    absorb="both",
    max_bond=None,
    cutoff=1e-10,
    cutoff_mode="rel",
    renorm=None,
    max_distance=1,
    smudge=1e-6,
    power=0.5,
    optimize=None,
    info=None,
):
    from quimb.tensor.decomp import svd

    t1, t2 = self._tids_get(tid1, tid2)

    # separate a local spanning tree into left and right components
    lset = oset([tid1])
    rset = oset([tid2])
    if max_distance > 0:
        for a, b, _ in self.get_tree_span(
            [tid1, tid2], max_distance=max_distance, inwards=False
        ):
            (lset if b in lset else rset).add(a)

    # for each tree, contract with dual to form reflected environment
    (bix,) = bonds(t1, t2)
    bix_dual = rand_uuid()

    tnl = self._select_tids(lset, virtual=False)
    tnr = self._select_tids(rset, virtual=False)

    tn_El = tnl & tnl.reindex({bix: bix_dual}).conj_()
    El = tn_El.to_dense([bix_dual], [bix], optimize=optimize)
    tn_Er = tnr & tnr.reindex({bix: bix_dual}).conj_()
    Er = tn_Er.to_dense([bix], [bix_dual], optimize=optimize)

    # decompose the environment matrices to use as a temporary gauge
    #     - this is like doing SVD(unreflected environment) but much
    #     cheaper since we only need the left singular vectors.
    sl2, Ul = do("linalg.eigh", El)
    sr2, Ur = do("linalg.eigh", Er)

    # we are free transform the singular values as inverse is just 1 / s
    sl = (sl2 + smudge * sl2[0]) ** power
    sr = (sr2 + smudge * sr2[0]) ** power
    sl = sl / do("max", sl)
    sr = sr / do("max", sr)

    # form the central reduced bond matrix & perform the actual truncation
    C = (reshape(sl, (-1, 1)) * dag(Ul)) @ (Ur * reshape(sr, (1, -1)))
    opts = _parse_split_opts(
        method, cutoff, absorb, max_bond, cutoff_mode, renorm
    )
    Lc, sc, Rc = svd(C, **opts)

    # combine the 'ungauge' with the central compressors and insert
    Cl = (Ul * reshape(1 / sl, (1, -1))) @ Lc
    Cr = Rc @ (reshape(1 / sr, (-1, 1)) * dag(Ur))

    t1.gate_(Cl.T, bix)
    t2.gate_(Cr, bix)

    if sc is not None and info is not None:
        info["singular_values"] = sc


def _compute_tensor_pair_env(
    self,
    tids,
    select_local_distance=None,
    select_local_opts=None,
    max_bond=None,
    cutoff=None,
    contract_around_distance=1,
    contract_around_opts=None,
    include=None,
    exclude=None,
):
    """Compute the local TN around ``tid``, maybe approximately."""
    # the TN we will start with
    if select_local_distance is include is exclude is None:
        # ... either the full TN
        tn_env = self.copy()

    else:
        # ... or just a local patch of the TN (with dangling bonds removed)
        select_local_opts = ensure_dict(select_local_opts)
        select_local_opts.setdefault("reduce_outer", "svd")

        tn_env = self._select_local_tids(
            tids,
            max_distance=select_local_distance,
            virtual=False,
            include=include,
            exclude=exclude,
            **select_local_opts,
        )

        # not propagated by _select_local_tids
        tn_env.exponent = self.exponent

    # possibly boundary contract in to make the env smaller
    if max_bond is not None:
        contract_around_opts = ensure_dict(contract_around_opts)
        tn_env._contract_around_tids(
            tids,
            max_bond=max_bond,
            cutoff=cutoff,
            min_distance=contract_around_distance,
            **contract_around_opts,
        )

    return tn_env


def _compress_between_full_gauge_tids(
    self,
    tid1,
    tid2,
    max_bond=None,
    cutoff=0.0,
    absorb="both",
    select_local_distance=None,
    select_local_opts=None,
    env_max_bond="max_bond",
    env_cutoff="cutoff",
    contract_around_distance=1,
    contract_around_opts=None,
    env_optimize="auto-hq",
    env_split_opts=None,
    smudge=1e-12,
    include=None,
    exclude=None,
    **compress_opts,
):
    r"""Use the full environment to try and temporarily gauge and compress.

    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┗━━━━━●━━━━━●━━━━━━━━U━s━VH━┛  (1) form and split the env
        ┊  1  ┊  2  ┊     ┈┈┈┈┈┈
        lix   bix   rix     env

    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┗Gi━G━●━━━━━●━━G━Gi━━U━s━VH━┛  (2) gauge the tensors using split
        ->       <-

    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┗━Gi━━○━━━━━○━━Gi━━━━U━s━VH━┛   (3) compress the gauged tensors
                *

    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┗━Gi━━○─────○━━Gi━━━━U━s━VH━┛   (4) ungauge the compressed tensors
        ->       <-

    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┗━━━━━●─────●━━━━━━━━U━s━VH━┛
                            ┈┈┈┈┈┈
                            env unchanged

    """
    if env_max_bond == "max_bond":
        env_max_bond = max_bond
    if env_cutoff == "cutoff":
        env_cutoff = cutoff

    ta = self.tensor_map[tid1]
    tb = self.tensor_map[tid2]

    tensor_fuse_squeeze(ta, tb)
    lix, bix, rix = group_inds(ta, tb)
    if not bix:
        return

    tn_env = self._compute_tensor_pair_env(
        (tid1, tid2),
        select_local_distance=select_local_distance,
        select_local_opts=select_local_opts,
        max_bond=env_max_bond,
        cutoff=env_cutoff,
        contract_around_distance=contract_around_distance,
        contract_around_opts=contract_around_opts,
        include=include,
        exclude=exclude,
    )

    # now we form an effective linear operator of the env
    tn_env = tn_env._select_without_tids((tid1, tid2))
    E = tn_env.aslinearoperator(rix, lix, optimize=env_optimize)

    # then we split it
    env_split_opts = ensure_dict(env_split_opts)
    env_split_opts.setdefault("cutoff", 0.0)
    U, s, VH = tensor_split(
        E,
        left_inds=lix,
        right_inds=rix,
        get="arrays",
        absorb=None,
        **env_split_opts,
    )

    # renormalize the env singular values for stability
    s = s / do("max", s) + smudge

    # create the gauges
    U = reshape(U, (-1, s.size))
    VH = reshape(VH, (s.size, -1))
    s_1_2 = s ** (1 / 2)
    s_1_2_inv = s ** (-1 / 2)
    gauge_a = reshape(s_1_2, (-1, 1)) * VH
    guage_a_inv = dag(VH) * reshape(s_1_2_inv, (1, -1))
    gauge_b = U * reshape(s_1_2, (1, -1))
    gauge_b_inv = reshape(s_1_2_inv, (-1, 1)) * dag(U)

    # gauge the raw arrays
    X_a = ta.to_dense(lix, bix)
    X_b = tb.to_dense(bix, rix)
    X_gauged_a = gauge_a @ X_a
    X_gauged_b = X_b @ gauge_b

    # perform the actual compression
    tga = Tensor(X_gauged_a, ("env_left_bond", "bond"))
    tgb = Tensor(X_gauged_b, ("bond", "env_right_bond"))
    tensor_compress_bond(
        tga, tgb, max_bond=max_bond, cutoff=cutoff, **compress_opts
    )

    # ungauge the compressed arrays
    X_compressed_a = guage_a_inv @ tga.data
    X_compressed_b = tgb.data @ gauge_b_inv

    # reshape the arrays to insert back into the original TN
    bix_size = tga.ind_size("bond")

    # create a tensor for the new compressed data, and update original ta
    tca_inds = lix + bix
    tca_shape = [bix_size if ix in bix else ta.ind_size(ix) for ix in tca_inds]
    tca = Tensor(reshape(X_compressed_a, tca_shape), tca_inds)
    tca.transpose_like_(ta)
    ta.modify(data=tca.data)

    # create a tensor for the new compressed data, and update original tb
    tcb_inds = bix + rix
    tcb_shape = [bix_size if ix in bix else tb.ind_size(ix) for ix in tcb_inds]
    tcb = Tensor(reshape(X_compressed_b, tcb_shape), tcb_inds)
    tcb.transpose_like_(tb)
    tb.modify(data=tcb.data)

    if absorb != "both":
        tensor_canonize_bond(ta, tb, absorb=absorb)


def gauge_product_boundary_vector(
    self,
    tags,
    which="all",
    max_bond=1,
    smudge=1e-6,
    canonize_distance=0,
    select_local_distance=None,
    select_local_opts=None,
    inplace=False,
    **contract_around_opts,
):
    # this is what we are going to eventually gauge
    tn = self if inplace else self.copy()
    tids = tn._get_tids_from_tags(tags, which)

    # form the double layer tensor network - this is the TN we will
    #     generate the actual gauges with
    if select_local_distance is None:
        # use the whole tensor network ...
        outer_inds = tn.outer_inds()
        dtn = tn.H & tn
    else:
        # ... or just a local patch
        select_local_opts = ensure_dict(select_local_opts)
        ltn = tn._select_local_tids(
            tids,
            max_distance=select_local_distance,
            virtual=False,
            **select_local_opts,
        )
        outer_inds = ltn.outer_inds()
        dtn = ltn.H | ltn

    # get all inds in the tagged region
    region_inds = set.union(*(set(tn.tensor_map[tid].inds) for tid in tids))

    # contract all 'physical' indices so that we have a single layer TN
    #     outside region and double layer sandwich inside region
    for ix in outer_inds:
        if (ix in region_inds) or (ix not in dtn.ind_map):
            # 1st condition - don't contract region sandwich
            # 2nd condition - if local selecting, will get multibonds so
            #     some indices already contracted
            continue
        dtn.contract_ind(ix)

    # form the single layer boundary of double layer tagged region
    dtids = dtn._get_tids_from_tags(tags, which)
    dtn._contract_around_tids(
        dtids,
        min_distance=1,
        max_bond=max_bond,
        canonize_distance=canonize_distance,
        **contract_around_opts,
    )

    # select this boundary and compress to ensure it is a product operator
    dtn = dtn._select_without_tids(dtids, virtual=True)
    dtn.compress_all_(max_bond=1)
    dtn.squeeze_()

    # each tensor in the boundary should now have exactly two inds
    #     connecting to the top and bottom of the tagged region double
    #     layer. Iterate over these, inserting the gauge into the original
    #     tensor network that would turn each of these boundary tensors
    #     into identities.
    for t in dtn:
        (ix,) = [i for i in t.inds if i in region_inds]
        _, s, VH = do("linalg.svd", t.data)
        s = s + smudge * s[0]
        G = reshape(s**0.5, (-1, 1)) * VH
        Ginv = dag(VH) * reshape(s**-0.5, (1, -1))

        tid_l, tid_r = sorted(tn.ind_map[ix], key=lambda tid: tid in tids)
        tn.tensor_map[tid_l].gate_(Ginv.T, ix)
        tn.tensor_map[tid_r].gate_(G, ix)

    return tn


def _compress_between_virtual_dense_tids(
    self,
    tidl,
    tidr,
    max_bond,
    cutoff,
    r,
    absorb="both",
    include=None,
    exclude=None,
    span_opts=None,
    contract_opts=None,
    optimize="auto-hq",
    **compress_opts,
):
    check_opt("absorb", absorb, ("both",))

    tl = self.tensor_map[tidl]
    tr = self.tensor_map[tidr]
    _, bix, _ = tensor_make_single_bond(tl, tr)

    # get a local region, but divide into left and right parts
    span_opts = ensure_dict(span_opts)
    span_opts["max_distance"] = r
    span_opts["exclude"] = exclude
    span_opts["include"] = include
    span_opts["inwards"] = False
    tree = self.get_tree_span([tidl, tidr], **span_opts)

    tidsl = oset([tidl])
    tidsr = oset([tidr])
    for tid_outer, tid_inner, _ in tree:
        if tid_inner in tidsl:
            tidsl.add(tid_outer)
        else:
            tidsr.add(tid_outer)

    # contract the squared region factors
    contract_opts = ensure_dict(contract_opts)
    contract_opts["optimize"] = optimize

    # compute reduced factor via XdagX for the left environment
    tnl = self._select_tids(tidsl, virtual=False)
    dl = self.inds_size(ix for ix in tnl.outer_inds() if ix != bix)
    dr = self.ind_size(bix)
    tn_XX = tnl.H.reindex({bix: "__conjbix__"}) & tnl
    XX = tn_XX.to_dense(["__conjbix__"], [bix], **contract_opts)
    Rl = decomp.squared_op_to_reduced_factor(XX, dl, dr)

    # compute reduced factor via XdagX for the right environment
    tnr = self._select_tids(tidsr, virtual=False)
    dl = self.inds_size(ix for ix in tnr.outer_inds() if ix != bix)
    dr = self.ind_size(bix)
    tn_XX = tnr.H.reindex({bix: "__conjbix__"}) & tnr
    XX = tn_XX.to_dense(["__conjbix__"], [bix], **contract_opts)
    Rr = decomp.squared_op_to_reduced_factor(XX, dl, dr)

    compress_opts["max_bond"] = max_bond
    compress_opts["cutoff"] = cutoff

    # compute the oblique projectors from the reduced factors
    Pl, Pr = decomp.compute_oblique_projectors(Rl, Rr.T, **compress_opts)

    # absorb the projectors into the tensors to perform the compression
    tl.gate_(Pl.T, bix)
    tr.gate_(Pr, bix)


def _compress_between_virtual_duotree_tids(
    self,
    tidl,
    tidr,
    max_bond,
    cutoff,
    r,
    absorb="both",
    include=None,
    exclude=None,
    span_opts=None,
    duo_exclude="tensor",
    **compress_opts,
):
    check_opt("absorb", absorb, ("both",))
    check_opt("duo_exclude", duo_exclude, ("tensor", "bond"))

    span_opts = ensure_dict(span_opts)
    span_opts["max_distance"] = r
    span_opts["include"] = include
    compress_opts["max_bond"] = max_bond
    compress_opts["cutoff"] = cutoff

    tl = self.tensor_map[tidl]
    tr = self.tensor_map[tidr]
    _, bix, _ = tensor_make_single_bond(tl, tr)

    if duo_exclude == "bond":
        # only exclude the actual shared bond from gauging
        tntree = self.copy()
        tntree._cut_between_tids(tidl, tidr, "__templ__", "__tempr__")
        treel = tntree.get_tree_span([tidl], exclude=exclude, **span_opts)
        treer = tntree.get_tree_span([tidr], exclude=exclude, **span_opts)

    else:  # duo_exclude == "tensor"
        # only exclude the other tensor from gauging
        if exclude is None:
            exclude_left = (tidr,)
            exclude_right = (tidl,)
        else:
            if not isinstance(exclude, oset):
                exclude = oset(exclude)
            exclude_left = exclude | oset([tidr])
            exclude_right = exclude | oset([tidl])

        treel = self.get_tree_span([tidl], exclude=exclude_left, **span_opts)
        treer = self.get_tree_span([tidr], exclude=exclude_right, **span_opts)

    # compute reduced factors from trees
    (Rl,) = self._compute_tree_gauges(treel, [(tidl, bix)])
    (Rr,) = self._compute_tree_gauges(treer, [(tidr, bix)])
    Pl, Pr = decomp.compute_oblique_projectors(Rl, Rr.T, **compress_opts)
    tl.gate_(Pl.T, bix)
    tr.gate_(Pr, bix)


def contract_greedily(
    self,
    max_score=0,
    max_contractions=float("inf"),
    cands=None,
    inplace=False,
):
    import heapq

    tn = self if inplace else self.copy()

    def _check_ind(ind):
        tids = tn.ind_map[ind]
        if len(tids) != 2:
            return
        ta, tb = tn._tids_get(*tids)
        sizea, sizeb = ta.size, tb.size
        oinds = tn.compute_contracted_inds(*tids)
        sizeab = tn.inds_size(oinds)
        score = sizeab - sizea - sizeb
        heapq.heappush(cands, (score, *tids))

    if cands is None:
        cands = []

    if len(cands) == 0:
        for ind in tn.ind_map:
            _check_ind(ind)

    c = 0
    while cands:
        # get the 'best' contraction
        score, tida, tidb = heapq.heappop(cands)

        if not (tida in tn.tensor_map and tidb in tn.tensor_map):
            # tensor a or b already contracted
            continue
        if score > max_score:
            # contraction isn't good enough, all others are also worse
            break

        # perform the contraction
        ta = tn.pop_tensor(tida)
        tb = tn.pop_tensor(tidb)
        tab = tensor_contract(ta, tb, preserve_tensor=True)
        tn |= tab

        # check how many contractions we've done
        c += 1
        if c >= max_contractions:
            break

        # add new potential candidates
        for ind in tab.inds:
            _check_ind(ind)

    return tn


contract_greedily_ = functools.partialmethod(contract_greedily, inplace=True)


def all_mode_renormalization():
    import numpy as np
    from quimb.tensor.decomp import lq_stabilized, qr_stabilized, rdmul, ldmul

    def make_random_truncator(d, chi, k=2):
        x = np.zeros((d, chi))
        np.fill_diagonal(x, 1)
        x[chi - k :, -k:] = (
            # np.random.randn(chi + k, k)
            np.random.choice([-1, 1], size=(d - chi + k, k))
        ) / k**0.5
        return x

    def svd_truncated_amr(
        x,
        cutoff=0.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
    ):
        assert cutoff == 0.0
        assert renorm == 0

        U, s, VH = np.linalg.svd(x, full_matrices=False)
        d = s.shape[0]
        ssqrt = s**0.5

        if max_bond < d:
            N = make_random_truncator(d, max_bond)
            U = rdmul(U, ssqrt) @ N
            VH = dag(N) @ ldmul(ssqrt, VH)
        else:
            U = rdmul(U, ssqrt)
            VH = ldmul(ssqrt, VH)

        if absorb == 0:
            return U, None, VH

        if absorb == -1:
            L, _, Q = lq_stabilized(VH)
            return U @ L, None, Q

        if absorb == 1:
            Q, _, R = qr_stabilized(U)
            return Q, None, R @ VH



from .builder import SparseOperatorBuilder
from .hilbertspace import HilbertSpace, parse_edges_to_unique


def make_edge_factory(coeff):
    """Ensure `coeff` is a function that takes two sites and returns an edge
    coeff."""
    if isinstance(coeff, dict):

        def edge_factory(cooa, coob):
            try:
                return coeff[(cooa, coob)]
            except KeyError:
                return coeff[(coob, cooa)]

    elif callable(coeff):
        edge_factory = coeff

    else:

        def edge_factory(cooa, coob):
            # constant
            return coeff

    return edge_factory


def make_node_factory(coeff):
    """Ensure `coeff` is a function that takes a site and returns a node
    coeff."""
    if isinstance(coeff, dict):

        def node_factory(coo):
            return coeff[coo]

    elif callable(coeff):
        node_factory = coeff

    else:

        def node_factory(coo):
            # constant
            return coeff

    return node_factory


def heisenberg_from_edges(
    edges,
    j=1.0,
    b=0.0,
    order=None,
    sector=None,
    symmetry=None,
    hilbert_space=None,
    dtype=None,
):
    r"""Create a Heisenberg Hamiltonian on the graph defined by ``edges``.

    .. math::

        H =
        \sum_{\{i,j\}}^{|E|}
        \left(
        J_x S^x_i S^x_j + J_y S^y_i S^y_j + J_z S^z_i S^z_j
        \right)
        +
        \sum_{i}^{|V|}
        \left(
        B_x S^x_i + B_y S^y_i + B_z S^z_i
        \right)

    where :math:`\{i,j\}` are the edges of the graph, and :math:`S^x_i` is the
    spin-1/2 operator acting on site :math:`i` in the x-direction, etc. Note
    positive values of :math:`J` correspond to antiferromagnetic coupling here,
    and the magnetic field is in the z-direction by default.

    Parameters
    ----------
    edges : Iterable[tuple[hashable, hashable]]
        The edges, as pairs of hashable 'sites', that define the graph.
        Multiple edges are allowed, and will be treated as a single edge.
    j : float or tuple[float, float, float] or dict or callable, optional
        The Heisenberg exchange coupling constant(s). If a single float is
        given, it is used for all three terms. If a tuple of three floats is
        given, they are used for the xx, yy, and zz terms respectively. Note
        that positive values of ``j`` correspond to antiferromagnetic coupling.
        A dict or callable can be supplied to have edge-dependent couplings.
    b : float or tuple[float, float, float] or dict or callable, optional
        The magnetic field strength(s). If a single float is given, it is used
        taken as a z-field. If a tuple of three floats is given, they are used
        for the x, y, and z fields respectively. A dict or callable can be
        supplied to have site-dependent fields.
    order : callable or sequence of hashable objects, optional
        If provided, use this to order the sites. If a callable, it should be a
        sorting key. If a sequence, it should be a permutation of the sites,
        and ``key=order.index`` will be used.
    sector : {None, str, int, ((int, int), (int, int))}, optional
        The sector of the Hilbert space. If None, no sector is assumed.
    symmetry : {None, "Z2", "U1", "U1U1"}, optional
        The symmetry of the Hilbert space if any. If `None` and a `sector` is
        provided, the symmetry will be inferred from the sector if possible.
    hilbert_space : HilbertSpace, optional
        The Hilbert space to use. If not given, one will be constructed
        automatically from the edges. This overrides the ``order``,
        ``symmetry``, and ``sector`` parameters.
    dtype : {None, str, type}, optional
        The data type of the Hamiltonian. If None, a default dtype will be
        used, np.float64 for real and np.complex128 for complex.

    Returns
    -------
    H : SparseOperatorBuilder
        The Hamiltonian as a SparseOperatorBuilder object.
    """
    sites, edges = parse_edges_to_unique(edges)

    if hilbert_space is None:
        hilbert_space = HilbertSpace(
            sites=sites,
            order=order,
            sector=sector,
            symmetry=symmetry,
        )

    H = SparseOperatorBuilder(hilbert_space=hilbert_space, dtype=dtype)

    j_factory = make_edge_factory(j)
    b_factory = make_node_factory(b)

    for cooa, coob in edges:
        # get possibly edge-dependent couplings
        jab = j_factory(cooa, coob)
        try:
            jx, jy, jz = jab
        except TypeError:
            jx, jy, jz = jab, jab, jab

        if jx == jy:
            # keep things real
            H += jx / 2, ("+", cooa), ("-", coob)
            H += jx / 2, ("-", cooa), ("+", coob)
        else:
            H += jx, ("sx", cooa), ("sx", coob)
            H += jy, ("sy", cooa), ("sy", coob)

        H += jz, ("sz", cooa), ("sz", coob)

    for site in sites:
        # get possibly site-dependent field
        b = b_factory(site)
        try:
            bx, by, bz = b
        except TypeError:
            bx, by, bz = 0, 0, b
        H += bx, ("sx", site)
        H += by, ("sy", site)
        H += bz, ("sz", site)

    return H


def fermi_hubbard_from_edges(
    edges,
    t=1.0,
    U=1.0,
    mu=0.0,
    order=None,
    sector=None,
    symmetry=None,
    hilbert_space=None,
    dtype=None,
    pauli_decompose=False,
):
    r"""Create a Fermi-Hubbard Hamiltonian on the graph defined by ``edges``.
    The Hamiltonian is given by:

    .. math::

        H =
        -t
        \sum_{\{i,j\}}^{|E|}
        \sum_{\sigma \in \uparrow, \downarrow}
        \left(
        c_{\sigma,i}^\dagger c_{\sigma,j} +
        c_{\sigma,j}^\dagger c_{\sigma,i}
        \right)
        +
        U
        \sum_{i}^{|V|}
        n_{\uparrow,i} n_{\downarrow,i}
        -
        \mu
        \sum_{i}^{|V|}
        \left(
        n_{\uparrow,i} + n_{\downarrow,i}
        \right)

    where :math:`\{i,j\}` are the edges of the graph, and :math:`c_{\sigma,i}`
    is the fermionic annihilation operator acting on site :math:`i` with spin
    :math:`\sigma`. The Jordan-Wigner transformation is used to implement
    fermionic statistics.

    Parameters
    ----------
    edges : Iterable[tuple[hashable, hashable]]
        The edges, as pairs of hashable 'sites', that define the graph.
        Multiple edges are allowed, but will be treated as a single edge.
    t : float or tuple[float, float] or dict or callable, optional
        The hopping amplitude. Default is 1.0. If a tuple it specifies the up
        and down spin hoppings respectively. A dict or callable can be supplied
        to have edge-dependent hoppings.
    U : float or dict or callable, optional
        The on-site interaction strength. Default is 1.0. A dict or callable
        can be supplied to have site-dependent interactions.
    mu : float or tuple[float, float] or dict or callable, optional
        The chemical potential. Default is 0.0. If a tuple it specifies the up
        and down spin chemical potentials respectively. A dict or callable can
        be supplied to have site-dependent chemical potentials.
    order : callable or sequence of hashable objects, optional
        If provided, use this to order the sites. If a callable, it should be a
        sorting key. If a sequence, it should be a permutation of the sites,
        and ``key=order.index`` will be used.
    sector : {None, str, int, ((int, int), (int, int))}, optional
        The sector of the Hilbert space. If None, no sector is assumed.
    symmetry : {None, "Z2", "U1", "U1U1"}, optional
        The symmetry of the Hilbert space if any. If `None` and a `sector` is
        provided, the symmetry will be inferred from the sector if possible.
    hilbert_space : HilbertSpace, optional
        The Hilbert space to use. If not given, one will be constructed
        automatically from the edges. This overrides the ``order``,
        ``symmetry``, and ``sector`` parameters.
    dtype : {None, str, type}, optional
        The data type of the Hamiltonian. If None, a default dtype will be
        used, np.float64 for real and np.complex128 for complex.
    pauli_decompose : bool, optional
        Whether to decompose the Hamiltonian into Pauli strings after
        Jordan-Wigner transforming. Default is False.

    Returns
    -------
    H : SparseOperatorBuilder
        The Hamiltonian as a SparseOperatorBuilder object.
    """
    sites, edges = parse_edges_to_unique(edges)

    if hilbert_space is None:
        hilbert_space = HilbertSpace(
            sites=[(s, coo) for s in "↑↓" for coo in sites],
            order=order,
            sector=sector,
            symmetry=symmetry,
        )

    H = SparseOperatorBuilder(
        hilbert_space=hilbert_space,
        dtype=dtype,
        jordan_wigner=True,
        pauli_decompose=pauli_decompose,
    )

    t_factory = make_edge_factory(t)
    U_factory = make_node_factory(U)
    mu_factory = make_node_factory(mu)

    for cooa, coob in edges:
        # hopping, possibly edge and spin-dependent
        t = t_factory(cooa, coob)
        try:
            t_u, t_d = t
        except TypeError:
            t_u, t_d = t, t
        H -= t_u, ("+", ("↑", cooa)), ("-", ("↑", coob))
        H -= t_u, ("+", ("↑", coob)), ("-", ("↑", cooa))
        H -= t_d, ("+", ("↓", cooa)), ("-", ("↓", coob))
        H -= t_d, ("+", ("↓", coob)), ("-", ("↓", cooa))

    for coo in sites:
        # interaction, possibly site-dependent
        U = U_factory(coo)
        H += U, ("n", ("↑", coo)), ("n", ("↓", coo))

        # chemical potential, possibly site and spin-dependent
        mu = mu_factory(coo)
        try:
            mu_u, mu_d = mu
        except TypeError:
            mu_u, mu_d = mu, mu
        H -= mu_u, ("n", ("↑", coo))
        H -= mu_d, ("n", ("↓", coo))

    return H


def fermi_hubbard_spinless_from_edges(
    edges,
    t=1.0,
    V=0.0,
    mu=0.0,
    delta=0.0,
    order=None,
    sector=None,
    symmetry=None,
    hilbert_space=None,
    dtype=None,
    pauli_decompose=False,
):
    r"""Create a spinless Fermi-Hubbard Hamiltonian on the graph defined by
    ``edges``. The Hamiltonian is given by:

    .. math::

        H =
        -t \sum_{\{i,j\}}^{|E|}
        \left( c_i^\dagger c_j + c_j^\dagger c_i \right)
        +
        V \sum_{\{i,j\}}^{|E|} n_i n_j
        -
        \mu \sum_{i}^{|V|} n_i
        +
        \Delta \sum_{\{i,j\}}^{|E|}
        \left( c_i^\dagger c_j^\dagger + c_j c_i \right)

    where :math:`\{i,j\}` are the edges of the graph, and :math:`c_i` is the
    fermionic annihilation operator acting on site :math:`i`. The Jordan-Wigner
    transformation is used to implement fermionic statistics.

    Parameters
    ----------
    edges : Iterable[tuple[hashable, hashable]]
        The edges, as pairs of hashable 'sites', that define the graph.
        Multiple edges are allowed, but will be treated as a single edge.
    t : float or dict or callable, optional
        The hopping amplitude. Default is 1.0. A dict or callable can be
        supplied to have edge-dependent hoppings.
    V : float or dict or callable, optional
        The nearest neighbor interaction strength. Default is 0.0. A dict or
        callable can be supplied to have edge-dependent interactions.
    mu : float or dict or callable, optional
        The chemical potential. Default is 0.0. A dict or callable can be
        supplied to have site-dependent chemical potentials.
    delta : float or dict or callable, optional
        The superconducting pairing strength. Default is 0.0. A dict or
        callable can be supplied to have edge-dependent pairings.
    order : callable or sequence of hashable objects, optional
        If provided, use this to order the sites. If a callable, it should be a
        sorting key. If a sequence, it should be a permutation of the sites,
        and ``key=order.index`` will be used.
    sector : {None, str, int, ((int, int), (int, int))}, optional
        The sector of the Hilbert space. If None, no sector is assumed.
    symmetry : {None, "Z2", "U1", "U1U1"}, optional
        The symmetry of the Hilbert space if any. If `None` and a `sector` is
        provided, the symmetry will be inferred from the sector if possible.
    hilbert_space : HilbertSpace, optional
        The Hilbert space to use. If not given, one will be constructed
        automatically from the edges. This overrides the ``order``,
        ``symmetry``, and ``sector`` parameters.
    dtype : {None, str, type}, optional
        The data type of the Hamiltonian. If None, a default dtype will be
        used, np.float64 for real and np.complex128 for complex.
    pauli_decompose : bool, optional
        Whether to decompose the Hamiltonian into Pauli strings after
        Jordan-Wigner transforming. Default is False.

    Returns
    -------
    H : SparseOperatorBuilder
        The Hamiltonian as a SparseOperatorBuilder object.
    """
    sites, edges = parse_edges_to_unique(edges)

    if hilbert_space is None:
        hilbert_space = HilbertSpace(
            sites=sites,
            order=order,
            sector=sector,
            symmetry=symmetry,
        )

    H = SparseOperatorBuilder(
        hilbert_space=hilbert_space,
        dtype=dtype,
        jordan_wigner=True,
        pauli_decompose=pauli_decompose,
    )

    t_factory = make_edge_factory(t)
    V_factory = make_edge_factory(V)
    mu_factory = make_node_factory(mu)
    delta_factory = make_edge_factory(delta)

    for cooa, coob in edges:
        # hopping, possibly edge-dependent
        t = t_factory(cooa, coob)
        H -= t, ("+", cooa), ("-", coob)
        H -= t, ("+", coob), ("-", cooa)

        # nearest neighbor interaction, possibly edge-dependent
        V = V_factory(cooa, coob)
        H += V, ("n", cooa), ("n", coob)

        # pairing, possibly edge-dependent
        delta = delta_factory(cooa, coob)
        H += delta, ("+", cooa), ("+", coob)
        H += delta, ("-", coob), ("-", cooa)

    # chemical potential, possibly site-dependent
    for coo in sites:
        mu = mu_factory(coo)
        H -= mu, ("n", coo)

    return H


def rand_operator(n, m, k, kmin=None, seed=None, ops="XYZ"):
    """Generate a random operator with n qubits and m terms.
    Each term is a sum of k operators acting on different qubits.
    The operators are chosen randomly from the set {X, Y, Z, +, -, n}.
    The coefficients are drawn from a normal distribution.

    Parameters
    ----------
    n : int
        The number of qubits.
    m : int
        The number of terms in the operator.
    k : int
        The number of operators in each term.
    kmin : int, optional
        The minimum number of operators in each term. If not given, kmin = k.
    seed : int, optional
        The random seed for reproducibility.
    ops : str, optional
        The set of operators to choose from.

    Returns
    -------
    SparseOperatorBuilder
        The random operator as a SparseOperatorBuilder object.
    """
    import numpy as np

    terms = []

    rng = np.random.default_rng(seed)
    allowed_ops = np.array(list(ops))

    if kmin is None:
        kmin = k
    if not (0 <= kmin <= k <= n):
        raise ValueError(
            "kmin must be positive and k must be between kmin and n"
        )

    for _ in range(m):
        coeff = rng.normal()
        ops = []

        if kmin == k:
            ki = k
        else:
            ki = rng.integers(kmin, k + 1)

        regs = rng.choice(np.arange(n), size=ki, replace=False)
        for reg in regs:
            op = rng.choice(allowed_ops)
            ops.append((str(op), int(reg)))
        terms.append((coeff, *ops))

    hilbert_space = HilbertSpace(sites=range(n))

    return SparseOperatorBuilder(terms=terms, hilbert_space=hilbert_space)
